local Bitmap = import('/lua/maui/bitmap.lua').Bitmap
local Group = import('/lua/maui/group.lua').Group
local UIUtil = import('/lua/ui/uiutil.lua')
local LayoutHelpers = import('/lua/maui/layouthelpers.lua')
local Window = import('/lua/maui/window.lua').Window
local Grid = import('/lua/maui/grid.lua').Grid
local buildingTemplates = import('/mods/build patterns/buildingtemplates.lua').BuildingTemplates
local Patterns = import('/mods/build patterns/patterntemplates.lua').Templates
local Decal = import('/lua/user/userdecal.lua').UserDecal
local ToolTip = import('/lua/ui/game/tooltip.lua')

--the UI elements
local background = false
local buttonGrid = false
local gridWidth = 72
local gridHeight = 72
--table of footprints, so they can all be moved and destroyed in the worldview handleevent
local buildFootprints = {}
--disables the buttons so it's impossible to give 2 orders at once
local Building = false
--tooltips
local mouseover = false
--rotation of the pattern
local rotation = 1
local rotationTable = {
	{1,1},
	{1,-1},
	{-1,-1},
	{-1,1},
}
local RadialBG = false

function Init()
	--create right hand screen menu
	background = Group(GetFrame(0))
	background.Top:Set(100)
	background.Height:Set(100)
	background.Right:Set(GetFrame(0).Right)
	background.Left:Set(function() return background.Right() - 10 end)
	background.Depth:Set(5000)
	background.bg = Group(background)
	LayoutHelpers.FillParent(background.bg, background)
	background.bg:DisableHitTest()
	background.bgTop = Bitmap(background.bg)
	background.bgMiddle = Bitmap(background.bg)
	background.bgBottom = Bitmap(background.bg)
	background.bgTop:SetTexture('/mods/build patterns/textures/backgrounds/bgtop.dds')
	background.bgMiddle:SetTexture('/mods/build patterns/textures/backgrounds/bgmiddle.dds')
	background.bgBottom:SetTexture('/mods/build patterns/textures/backgrounds/bgbottom.dds')
	LayoutHelpers.AtLeftTopIn(background.bgTop, background)
	LayoutHelpers.AtLeftIn(background.bgMiddle, background)
	LayoutHelpers.AtLeftIn(background.bgBottom, background)
	LayoutHelpers.AtBottomIn(background.bgBottom, background)
	background.bgMiddle.Top:Set(background.bgTop.Bottom)
	background.bgMiddle.Bottom:Set(background.bgBottom.Top)
	
	buttonGrid = Grid(background, gridWidth, gridHeight)
	LayoutHelpers.AtLeftTopIn(buttonGrid, background, 8, 8)
	buttonGrid.Right:Set(function() return buttonGrid.Left() + gridWidth end)
	buttonGrid.Bottom:Set(function() return buttonGrid.Top() + gridHeight end)
	buttonGrid.Depth:Set(background.Depth() + 10)
	background.HandleEvent = function(self, event)
		if event.Type == 'MouseEnter' then
			if self:GetAlpha() < 0.7 then
				Fade(self, 0.5, 1)
			else
				self.bg:SetNeedsFrameUpdate(false)
				self.bg:SetAlpha(1, true)
				self.Left:Set(function() return self.Right() - 87 end)
			end
		elseif event.Type == 'MouseExit' then
			Fade(self, 0.5, 0.3, 0.15)
			CloseToolTip()
		end
	end
	
	background.Height:Set(function() return buttonGrid.Bottom() - buttonGrid.Top() + 16 end)
	background:SetAlpha(0.2, true)
	
	--create radial menu
	local GameParent = import('/lua/ui/game/gamemain.lua').GetGameParent()
	local oldHandleEvent = GameParent.HandleEvent
	GameParent.HandleEvent = function(self, event)
		if event.Type=="ButtonPress" then
			if RadialBG then
				if event.Modifiers.Middle then 
					if event.Modifiers.Shift == true then
						RadialBG.Left:Set(event.MouseX-80)
						RadialBG.Top:Set(event.MouseY-80)
						RadialBG:Show()
					else
						RadialBG:Hide()
					end
				else
					RadialBG:Hide()
				end
			end
		end
		oldHandleEvent(self,event)
	end
end

--effect function to fade and slide the panel in/out
function Fade(control, time, finalValue, delay)
	control.bg:SetNeedsFrameUpdate(true)

	local initAlpha = control.bg:GetAlpha()
	local finalAlpha =  finalValue
	local timeAccum = 0
	local wait = delay or 0
	
	control.bg.OnFrame = function(self, elapsedTime)
		timeAccum = timeAccum + elapsedTime
		if timeAccum > wait then
			self:SetAlpha(math.min(MATH_Lerp(timeAccum - wait, 0, time, initAlpha, finalAlpha), 1), true)
			control.Left:Set(function() return control.Right() - self:GetAlpha()*110 + 23 end)
			if timeAccum >= time + wait then
				self:SetNeedsFrameUpdate(false)
				self:SetAlpha(finalAlpha, true)
				control.Left:Set(function() return control.Right() - finalAlpha*110 + 23 end)
			end
		end
	end
end

local radialMenuButtonPositions = {
	{0, -56},
	{56, -56},
	{56, 0},
	{56, 56},
	{0, 56},
	{-56, 56},
	{-56, 0},
	{-56, -56},
}

--create the radial menu bg
function CreateRadialGroup(faction, buildable)
	if not RadialBG then
		RadialBG = Bitmap(GetFrame(0), '/mods/Build Patterns/textures/radial_backgrounds/radialbg_' .. faction .. '.dds')
		RadialBG:SetAlpha(0.5)
		RadialBG.Depth:Set(1000)
		RadialBG.Left:Set(400)
		RadialBG.Top:Set(400)
		RadialBG.HandleEvent = function(self, event)
			if event.Type == 'ButtonPress' then
				RadialBG:Hide()
				return true
			end
		end
		local index = 1
		for name, data in Patterns do
			--only show buttons for buildable patterns
			for i, id in buildable do
				if index <= 8 then
					if id == buildingTemplates[faction][data.FirstBuilding] then
						local button = CreateButton(name, faction, RadialBG)
						LayoutHelpers.AtCenterIn(button, RadialBG, radialMenuButtonPositions[index][1], radialMenuButtonPositions[index][2])
						index = index + 1
					end
				end
			end
		end
		RadialBG:Hide()
	end
end

--create the construction button
local UPPER_GLOW_THRESHHOLD = .5
local LOWER_GLOW_THRESHHOLD = .1
local GLOW_SPEED = 2

function CreateButton(name, Faction, parent)
	local btn = Bitmap(parent)
	btn:SetTexture('/mods/build patterns/textures/buttons' ..Faction.. '/' ..name.. '.dds')
	if parent == RadialBG then
		btn.Width:Set(54)
		btn.Height:Set(54)
	else
		btn.Width:Set(gridWidth - 2)
		btn.Height:Set(gridHeight - 2)
	end
	
	--create the flashy glow when over a button
	btn.Glow = Bitmap(btn)
	btn.Glow:SetTexture(UIUtil.UIFile('/game/units_bmp/glow.dds'))
	LayoutHelpers.FillParent(btn.Glow, btn)
	btn.Glow:SetAlpha(0)
	btn.Incrementing = true
	btn.Glow.OnFrame = function(self, elapsedTime)
		local curAlpha = self:GetAlpha()
		if self.Incrementing then
			curAlpha = curAlpha + (elapsedTime * GLOW_SPEED)
			if curAlpha > UPPER_GLOW_THRESHHOLD then
				curAlpha = UPPER_GLOW_THRESHHOLD
				self.Incrementing = false
			end
		else
			curAlpha = curAlpha - (elapsedTime * GLOW_SPEED)
			if curAlpha < LOWER_GLOW_THRESHHOLD then
				curAlpha = LOWER_GLOW_THRESHHOLD
				self.Incrementing = true
			end
		end
		self:SetAlpha(curAlpha)
	end

	btn.HandleEvent = function(self, event)
		if not Building then
			local worldview = import('/lua/ui/game/worldview.lua').viewLeft
			local oldHandleEvent = worldview.HandleEvent
			if event.Type == 'ButtonPress' then
				local sound = Sound({Cue = 'UI_Opt_Mini_Button_Click', Bank = "Interface",})
				PlaySound(sound)
				self.Glow:SetNeedsFrameUpdate(false)
				self.Glow:SetAlpha(0)
				local selection = GetSelectedUnits()
				if selection then
					--get the first building, if for some reason it's not in the conversion table (name to ID), then do nothing
					local firstBuild = buildingTemplates[Faction][Patterns[name].FirstBuilding]
					if firstBuild then
						--building active, stops a second build order being given before this one is ended
						Building = true
						--if a command is issued, start the thread to close down the pattern ui stuff
						local issuedCommandThread = false
						--start command mode for the first building, and fork a thread to show all the secondary ones
						import('/lua/ui/game/commandmode.lua').StartCommandMode("build", {name = firstBuild})
						local showBuildingThread = ForkThread(showBuildings, {name = name, Faction = Faction})
						worldview.HandleEvent = function(self2, event2)
							--move the secondary footprints around with the mouse
							for i, v in buildFootprints do
								local mousepos = GetMouseWorldPos()
								--buildingpos is the world position of the building, it's slighly complex so that only one offset table has to be provided to deal with all rotations.
								--math.floor is to snap buildings to the nearest buildable position (0.5 is because buildings are placed at e.g. 1.5 not 1)
								--rotationTable[stuff] is just a 1 or -1 modifier to determine which direction an offset should go
								--math.mod(rotation, 2) + 1 returns either the x offset or the z offset
								local buildingpos = {
									math.floor(mousepos[1]+rotationTable[rotation][1]*v.offset[math.mod(rotation, 2) + 1])+0.5,
									mousepos[2],
									math.floor(mousepos[3]+rotationTable[rotation][2]*v.offset[math.mod(rotation + 1, 2) + 1])+0.5
								}
								local footprintScreenPos = worldview:Project(buildingpos)
								v.footprint:SetPositionByScreen({footprintScreenPos[1] + worldview.Left(), footprintScreenPos[2] + worldview.Top()})
							end
							if event2.Type == 'ButtonPress' then
								--start the build thread, and destroy all the ui stuff, and return the worldview event back to the old one
								if event2.Modifiers.Left then
									ForkThread(waitthread, {Name = name, Position = GetMouseWorldPos(), Faction = Faction, Rotation = rotation})
									if not issuedCommandThread then
										issuedCommandThread = ForkThread(IssuedCommandThread, {showBuildingThread = showBuildingThread, oldHandleEvent = oldHandleEvent, worldview = worldview})
									end
								elseif event2.Modifiers.Middle then
									rotation = rotation + 1
									if rotation > 4 then
										rotation = 1
									end
									return true
								elseif event2.Modifiers.Right then
									KillThread(showBuildingThread)
									for i, v in buildFootprints do
										v.footprint:Destroy()
										buildFootprints[i] = nil
									end
									Building = false
									worldview.HandleEvent = oldHandleEvent
								end
							end
						end
					end
				end
			elseif event.Type == 'MouseEnter' then
				local sound = Sound({Cue = 'UI_Opt_Mini_Button_Over', Bank = "Interface",})
				PlaySound(sound)
	            self.Glow:SetNeedsFrameUpdate(true)
	            self.Glow:SetAlpha(LOWER_GLOW_THRESHHOLD)
				CreateToolTip(self, {Title = name, Body = Patterns[name].Description})
			elseif event.Type == 'MouseExit' then
				self.Glow:SetNeedsFrameUpdate(false)
				self.Glow:SetAlpha(0)
				CloseToolTip()
			end
		end
	end
	return btn
end

--called by keymapping
function KeymapStart(name)
	if not Building then
		local worldview = import('/lua/ui/game/worldview.lua').viewLeft
		local oldHandleEvent = worldview.HandleEvent
		local selection = GetSelectedUnits()
		if selection then
			--get faction
			local Faction = 'AEON'
			for i, cat in selection[1]:GetBlueprint().Categories do
				if cat == 'CYBRAN' then
					Faction = 'CYBRAN'
				elseif cat == 'UEF' then
					Faction = 'UEF'
				end
			end
			--get the first building, if for some reason it's not in the conversion table (name to ID), then do nothing
			local firstBuild = buildingTemplates[Faction][Patterns[name].FirstBuilding]
			if firstBuild then
				--building active, stops a second build order being given before this one is ended
				Building = true
				--if a command is issued, start the thread to close down the pattern ui stuff
				local issuedCommandThread = false
				--start command mode for the first building, and fork a thread to show all the secondary ones
				import('/lua/ui/game/commandmode.lua').StartCommandMode("build", {name = firstBuild})
				local showBuildingThread = ForkThread(showBuildings, {name = name, Faction = Faction})
				worldview.HandleEvent = function(self2, event2)
					--move the secondary footprints around with the mouse
					for i, v in buildFootprints do
						local mousepos = GetMouseWorldPos()
						--buildingpos is the world position of the building, it's slighly complex so that only one offset table has to be provided to deal with all rotations.
						--math.floor is to snap buildings to the nearest buildable position (0.5 is because buildings are placed at e.g. 1.5 not 1)
						--rotationTable[stuff] is just a 1 or -1 modifier to determine which direction an offset should go
						--math.mod(rotation, 2) + 1 returns either the x offset or the z offset
						local buildingpos = {
							math.floor(mousepos[1]+rotationTable[rotation][1]*v.offset[math.mod(rotation, 2) + 1])+0.5,
							mousepos[2],
							math.floor(mousepos[3]+rotationTable[rotation][2]*v.offset[math.mod(rotation + 1, 2) + 1])+0.5
						}
						local footprintScreenPos = worldview:Project(buildingpos)
						v.footprint:SetPositionByScreen({footprintScreenPos[1] + worldview.Left(), footprintScreenPos[2] + worldview.Top()})
					end
					if event2.Type == 'ButtonPress' then
						--start the build thread, and destroy all the ui stuff, and return the worldview event back to the old one
						if event2.Modifiers.Left then
							ForkThread(waitthread, {Name = name, Position = GetMouseWorldPos(), Faction = Faction, Rotation = rotation})
							if not issuedCommandThread then
								issuedCommandThread = ForkThread(IssuedCommandThread, {showBuildingThread = showBuildingThread, oldHandleEvent = oldHandleEvent, worldview = worldview})
							end
						elseif event2.Modifiers.Middle then
							rotation = rotation + 1
							if rotation > 4 then
								rotation = 1
							end
							return true
						elseif event2.Modifiers.Right then
							KillThread(showBuildingThread)
							for i, v in buildFootprints do
								v.footprint:Destroy()
								buildFootprints[i] = nil
							end
							Building = false
							worldview.HandleEvent = oldHandleEvent
						end
					end
				end
			end
		end
	end
end

--set up the construction grid
function LayoutButtons(faction, buildable)
	buttonGrid:DeleteAndDestroyAll(true)
	local visCols, visRows = buttonGrid:GetVisible()
	local currentRow = 1
	local currentCol = 1
	buttonGrid:AppendCols(visCols, true)
	buttonGrid:AppendRows(1, true)
	local index = 0
	for name, data in Patterns do
		--only show buttons for buildable patterns
		for i, id in buildable do
			if id == buildingTemplates[faction][data.FirstBuilding] then
				index = index + 1
				local button = CreateButton(name, faction, buttonGrid)
				buttonGrid:SetItem(button, currentCol, currentRow, true)
				if currentCol < visCols then
					currentCol = currentCol + 1
				else
					currentCol = 1
					currentRow = currentRow + 1
					buttonGrid:AppendRows(1, true)
				end
			end
		end
	end
	buttonGrid.Bottom:Set(function() return buttonGrid.Top() + index * gridHeight end)
	buttonGrid:EndBatch()
end

function IssuedCommandThread(args)
	while Building do
		if not IsKeyDown('Shift') then
			KillThread(args.showBuildingThread)
			for i, v in buildFootprints do
				v.footprint:Destroy()
				buildFootprints[i] = nil
			end
			Building = false
			args.worldview.HandleEvent = args.oldHandleEvent
		end
		WaitSeconds(0.1)
	end
end

function showBuildings(args)
	local worldview = import('/lua/ui/game/worldview.lua').viewLeft
	while true do
		for i, building in Patterns[args.name].FollowingBuildings do
			local fadetime = table.getsize(Patterns[args.name].FollowingBuildings)/9 + 0.4
			local buildingId = buildingTemplates[args.Faction][building[1]]
			if buildingId then
				local mousepos = GetMouseWorldPos()
				local buildingpos = {
					math.floor(mousepos[1]+rotationTable[rotation][1]*building[2][math.mod(rotation, 2) + 1])+0.5,
					mousepos[2],
					math.floor(mousepos[3]+rotationTable[rotation][2]*building[2][math.mod(rotation+1, 2) + 1])+0.5
				}
				--set up the footprints (they are positioned in the worldview handleevent)
				if not buildFootprints[i] then
					local footprintSizeX = __blueprints[buildingId].Physics.SkirtSizeX
					local footprintSizeZ = __blueprints[buildingId].Physics.SkirtSizeZ
					buildFootprints[i] = {}
					buildFootprints[i].offset = {building[2][1], building[2][2]}
					buildFootprints[i].footprint = Decal(GetFrame(0))
					if footprintSizeX <= 1 then
						buildFootprints[i].footprint:SetTexture('/mods/build patterns/textures/footprints/footprint-1by1.dds')
					elseif footprintSizeX <= 2 then
						buildFootprints[i].footprint:SetTexture('/mods/build patterns/textures/footprints/footprint-2by2.dds')
					elseif footprintSizeX <= 4 then
						buildFootprints[i].footprint:SetTexture('/mods/build patterns/textures/footprints/footprint-4by4.dds')
					elseif footprintSizeX <= 6 then
						buildFootprints[i].footprint:SetTexture('/mods/build patterns/textures/footprints/footprint-6by6.dds')
					else
						buildFootprints[i].footprint:SetTexture('/mods/build patterns/textures/footprints/footprint-8by8.dds')
					end
					buildFootprints[i].footprint:SetScale({footprintSizeX, 0, footprintSizeZ})
				end
				--show the flashing blue building models
				AddCommandFeedbackBlip({
					Position = buildingpos, 
					BlueprintID = buildingId,			
					TextureName = '/meshes/game/flag02d_albedo.dds',
					ShaderName = 'CommandFeedback',
					UniformScale = 1,
				}, fadetime)
				WaitSeconds(0.07)
			end
		end
		WaitSeconds(0.6)
	end
end


function waitthread(args)
	--wait a moment before giving the commands, or they will all get cancelled (no idea why)
	WaitSeconds(0.3)
	SimCallback({ Func = 'BuildPattern',
		Args = {Name = args.Name,
			List = Patterns[args.Name].FollowingBuildings,
			Position = args.Position,
			Faction = args.Faction,
			Rotation = args.Rotation,
			},
		}, true)
end

--issue the secondary build commands
function Build(Data, Units)
	for i, building in Data.List do
		local buildingId = buildingTemplates[Data.Faction][building[1]]
		if buildingId then
			local xLoc = math.floor(Data.Position[1] + rotationTable[Data.Rotation][1]*building[2][math.mod(Data.Rotation, 2) + 1])+0.5
			local zLoc = math.floor(Data.Position[3] + rotationTable[Data.Rotation][2]*building[2][math.mod(Data.Rotation+1, 2) + 1])+0.5
			local yLoc = GetTerrainHeight(xLoc, zLoc)
			IssueBuildMobile(Units, {xLoc, yLoc, zLoc}, buildingId, {})
		end
	end
end

function SelectionChanged(selection)
	if RadialBG then RadialBG:Destroy() RadialBG = false end
	if table.getn(selection) == 0 then
		background:Hide()
	else
		local engineer = false
		local faction = 'UEF'
		local bp = selection[1]:GetBlueprint()
		for i, cat in bp.Categories do
			if cat == 'ENGINEER' then
				engineer = true
			elseif cat == 'UEF' then
				faction = 'UEF'
			elseif cat == 'AEON' then
				faction = 'AEON'
			elseif cat == 'CYBRAN' then
				faction = 'CYBRAN'
			end
		end
		if engineer then
			--get list of all units the current unit can build
			local availableOrders,  availableToggles, buildableCategories = GetUnitCommandData(selection)
			local buildable = EntityCategoryGetUnitList(buildableCategories)

			LayoutButtons(faction, buildable)
			CreateRadialGroup(faction, buildable)
			background:Show()
		else
			background:Hide()
		end
	end
end

--tooltip function so I don't have to position every tooltip manually
function CreateToolTip(control, data, togglebutton)
	if mouseover then mouseover:Destroy() mouseover = false end
	local text = data.Title
	local body = data.Body
	mouseover = ToolTip.CreateExtendedToolTip(control, text, body)
	mouseover.Top:Set(function() return control.Top() end)
	mouseover.Right:Set(function() return control.Left() - 4 end)
	mouseover:SetAlpha(0, true)
	mouseover:SetNeedsFrameUpdate(true)
	
	local delayTime = 0.6
	local totaltime = 0
	mouseover.OnFrame = function(self, deltaTime)
		if totaltime > delayTime then
			local alpha = self:GetAlpha() + (deltaTime * 4)
			if alpha < 1 then
				self:SetAlpha(alpha, true)
			else
				self:SetAlpha(1, true)
				self:SetNeedsFrameUpdate(false)
			end
		end
		totaltime = totaltime + deltaTime
	end
end
		
function CloseToolTip()
	if mouseover then mouseover:Destroy() mouseover = false end
end
